/******************************************************************************/
/*                                                                            */
/* Version 1.5                                                                */
/* Author: Arjan van Vught                 Date: October   2001               */
/*                                                                            */
/******************************************************************************/
/* Version 1.0   First release NT                                             */
/* Version 1.1   First release AIX                                            */
/* Version 1.2   Updated for MQSeries Integrator 2.0.1                        */
/* Version 1.3   First release Sun Solaris                                    */
/* Version 1.4   First release Linux (on Intel)                               */
/* Version 1.5   First release HP-UX                                          */
/******************************************************************************/
/*                                                                            */
/* Note: This plugin node implementation is not necessarily portable          */
/* to all platforms supported by MQSeries Integrator Version 2.               */
/*                                                                            */
/******************************************************************************/
/* Includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>
#include <ctype.h>

#include "MQSIEparser.h"

/******************************************************************************/
/*                                                                            */
/* Plugin Parser Utility Function: Advance to next character                  */
/*                                                                            */
/******************************************************************************/
void advance(PARSER_CONTEXT_ST *context, CciParser *parser)
{
   int rc = 0;

   /* Advance to the next character */
   context->iIndex++;

   /* Detect and handle the end condition */
   if (context->iIndex == context->iSize)
      return;
   /* Obtain the next character from the buffer */
   context->iCurrentCharacter = cpiBufferByte(&rc, parser, context->iIndex);

   return;
}


/******************************************************************************/
/*                                                                            */
/* Plugin Parser Utility Function: Advance to end of a valid name             */
/*                                                                            */
/******************************************************************************/
void goToNameEnd(PARSER_CONTEXT_ST *context, CciParser *parser)
{
   /* Advance until a non-name character is found */
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
	while (context->iCurrentCharacter != chOpen )
#else
   while (context->iCurrentCharacter != L'(' )
#endif
      advance(context, parser);

   return;  
}


/******************************************************************************/
/*                                                                            */
/* Plugin Parser Utility Function: Advance to end of a tag                    */
/*                                                                            */
/******************************************************************************/
void goToValueEnd(PARSER_CONTEXT_ST *context, CciParser *parser)
{
   /* Advance until a non-name character is found */
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
	while (context->iCurrentCharacter != chClose )
#else
   while (context->iCurrentCharacter != L')' )
#endif
      advance(context, parser);

   return;
}


/******************************************************************************/
/*                                                                            */
/* Plugin Parser Utility Function: Skip white space from current position     */
/*                                                                            */
/******************************************************************************/
void skipWhiteSpace(PARSER_CONTEXT_ST *context, CciParser *parser)
{
  
   /* Advance until end of white-space or end of buffer */
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
	while (CciCharIsSpace((CciChar)context->iCurrentCharacter)) {
#else
	while (iswspace((CciChar)context->iCurrentCharacter)) {
#endif
   /* while (context->iCurrentCharacter == L' ' ) { */
      advance(context, parser);
      if (context->iIndex == context->iSize)
         return;
   }

   return;
}


int WhatType(const CciChar *str)
{
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
	if (CciCharCmp((CciChar *)constDATATYPE, str ) == 0)
		return CCI_VALUE_TYPE_CHAR;
	if (CciCharCmp((CciChar *)constCLASS, str ) == 0)
		return CCI_VALUE_TYPE_CHAR;
	if (CciCharCmp((CciChar *)constCONTROLNUM, str ) == 0)
		return CCI_VALUE_TYPE_INTEGER;
	if (CciCharCmp((CciChar *)constMSGNAME, str ) == 0)
		return CCI_VALUE_TYPE_CHAR;
	if (CciCharCmp((CciChar *)constMSGSEQNO, str ) == 0)
		return CCI_VALUE_TYPE_CHAR;
	if (CciCharCmp((CciChar *)constRECLEN, str ) == 0)
		return CCI_VALUE_TYPE_INTEGER;
	if (CciCharCmp((CciChar *)constRECDLM, str ) == 0)
		return CCI_VALUE_TYPE_CHAR;
	if (CciCharCmp((CciChar *)constRECFM, str ) == 0)
		return CCI_VALUE_TYPE_CHAR;
	if (CciCharCmp((CciChar *)constSENDER, str ) == 0)
		return CCI_VALUE_TYPE_CHAR;
	if (CciCharCmp((CciChar *)constSENDQUAL, str ) == 0)
		return CCI_VALUE_TYPE_CHAR;
	if (CciCharCmp((CciChar *)constRECEIVER, str ) == 0)
		return CCI_VALUE_TYPE_CHAR;
	if (CciCharCmp((CciChar *)constRECVQUAL, str ) == 0)
		return CCI_VALUE_TYPE_CHAR;
	if (CciCharCmp((CciChar *)constFROM, str ) == 0)
		return CCI_VALUE_TYPE_CHAR;
	if (CciCharCmp((CciChar *)constUNIQUEID, str ) == 0)
		return CCI_VALUE_TYPE_CHAR;
	if (CciCharCmp((CciChar *)constTIME, str ) == 0)
		return CCI_VALUE_TYPE_DATE;
	if (CciCharCmp((CciChar *)constDATE, str ) == 0)
		return CCI_VALUE_TYPE_TIME;

	return CCI_VALUE_TYPE_CHAR;

#else
  int retVal = CCI_VALUE_TYPE_CHAR;
  int i      = 0;

  KeywordType Table[] = {
		{ L"DATATYPE",   CCI_VALUE_TYPE_CHAR },
		{ L"CLASS",      CCI_VALUE_TYPE_CHAR },
		{ L"CONTROLNUM", CCI_VALUE_TYPE_INTEGER },
		{ L"MSGNAME",    CCI_VALUE_TYPE_CHAR },
		{ L"MSGSEQNO",   CCI_VALUE_TYPE_CHAR },
		{ L"RECLEN",     CCI_VALUE_TYPE_INTEGER },
		{ L"RECDLM",     CCI_VALUE_TYPE_CHAR },
		{ L"RECFM",      CCI_VALUE_TYPE_CHAR },
		{ L"SENDER",     CCI_VALUE_TYPE_CHAR },
		{ L"SENDQUAL",   CCI_VALUE_TYPE_CHAR },
		{ L"RECEIVER",   CCI_VALUE_TYPE_CHAR },
		{ L"RECVQUAL",   CCI_VALUE_TYPE_CHAR },
		{ L"FROM",       CCI_VALUE_TYPE_CHAR },
		{ L"UNIQUEID",   CCI_VALUE_TYPE_CHAR },
		{ L"DATE",       CCI_VALUE_TYPE_DATE },
		{ L"TIME",       CCI_VALUE_TYPE_TIME },
		{ NULL, 0 }, 
   };

   while (Table[i].Keyword != NULL) {
      if (wcscmp(Table[i].Keyword, str ) == 0)
         return(Table[i].Type);
      i++;
   }

   return retVal;
#endif																								/* 21/07/2001 */

}


CciContext *cpiCreateContext(CciParser *parser)
{
   PARSER_CONTEXT_ST *pc = NULL;

#ifdef _DEBUG  
   DBGPRINTF((OST, "->cpiCreateContext()\n"));
#endif      

   /* Allocate a pointer to the local context */
   pc = (PARSER_CONTEXT_ST *)malloc(sizeof(PARSER_CONTEXT_ST));

   if (pc) {
      /* Clear the context area */
      memset(pc, 0, sizeof(PARSER_CONTEXT_ST));
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
		CciCharCpy(pc->NextParserClassName, constMQSIE);
#else
      wcscpy(pc->NextParserClassName, L"MQSIE");
#endif
   }
   else {
    /* Error: Unable to allocate memory */
   }

#ifdef _DEBUG  
   DBGPRINTF((OST, "<-cpiCreateContext()\n"));
#endif      

   /* Return the pointer to the local context */
   return((CciContext *)pc);
}


void cpiDeleteContext(CciParser *parser, CciContext *c)
{
   PARSER_CONTEXT_ST *pc = (PARSER_CONTEXT_ST*)c;

#ifdef _DEBUG  
   DBGPRINTF((OST, "->cpiDeleteContext()\n"));
#endif

   free(pc);

#ifdef _DEBUG  
   DBGPRINTF((OST, "<-cpiDeleteContext()\n"));
#endif

   return;
}


/******************************************************************************/
/*                                                                            */
/* Plugin Parser Implementation Function:                                     */
/*                                                                            */
/******************************************************************************/
void cpiParseFirstChild(CciParser *parser, CciContext *context,  CciElement *element)
{
   PARSER_CONTEXT_ST* pc = context;
   int                rc;

#ifdef _DEBUG  
   DBGPRINTF((OST,"->cpiParseFirstChild() parser = 0x%x, context = 0x%x, element = 0x%x\n", parser, context, element));
#endif  

   if ((!cpiElementCompleteNext(&rc, element)) && (cpiElementType(&rc, element) == CCI_ELEMENT_TYPE_NAME)) {
      while ((!cpiElementCompleteNext(&rc, element)) && (!cpiFirstChild(&rc, element)) && (pc->iCurrentElement))
         pc->iCurrentElement = parseNextItem(parser, context, pc->iCurrentElement);
   }

#ifdef _DEBUG  
   DBGPRINTF((OST,"<-cpiParseFirstChild()\n"));
#endif

   return;
}

/******************************************************************************/
/*                                                                            */
/* Plugin Parser Implementation Function:                                     */
/*                                                                            */
/******************************************************************************/
void cpiParseLastChild(CciParser *parser, CciContext *context, CciElement *element)
{
   PARSER_CONTEXT_ST* pc = context;
   int                rc;

#ifdef _DEBUG  
   DBGPRINTF((OST, "->cpiParseLastChild() parser = 0x%x, context = 0x%x, element = 0x%x\n", parser, context, element));
#endif
  
   if ((cpiElementType(&rc, element) == CCI_ELEMENT_TYPE_NAME)) {
      while ((!cpiElementCompleteNext(&rc, element)) && (pc->iCurrentElement))
         pc->iCurrentElement = parseNextItem(parser, context, pc->iCurrentElement);
   }

#ifdef _DEBUG  
   DBGPRINTF((OST, "<-cpiParseLastChild()\n"));
#endif

  return;
}

/******************************************************************************/
/*                                                                            */
/* Plugin Parser Implementation Function:                                     */
/*                                                                            */
/******************************************************************************/
void cpiParseLeftSibling(CciParser *parser, CciContext *context, CciElement *element)
{
#ifdef _DEBUG  
   DBGPRINTF((OST, "->cpiParseLeftSibling() parser = 0x%x, context = 0x%x, element = 0x%x\n", parser, context, element));
  
   DBGPRINTF((OST, "<-cpiParseLeftSibling()\n"));
#endif

   return;
}

/******************************************************************************/
/*                                                                            */
/* Plugin Parser Implementation Function:                                     */
/*                                                                            */
/******************************************************************************/
void cpiParseRightSibling(CciParser *parser, CciContext *context, CciElement *element)
{
   PARSER_CONTEXT_ST* pc = context;
   int                rc;

#ifdef _DEBUG  
   DBGPRINTF((OST, "->cpiParseRightSibling() parser = 0x%x, context = 0x%x, element = 0x%x\n", parser, context, element));
#endif
  
   while ((!cpiElementCompleteNext(&rc, cpiParent(&rc, element))) && (!cpiNextSibling(&rc, element)) && (pc->iCurrentElement))
      pc->iCurrentElement = parseNextItem(parser, context, pc->iCurrentElement);


#ifdef _DEBUG  
   DBGPRINTF((OST, "<-cpiParseRightSibling()\n"));
#endif

   return;
}


/******************************************************************************/
/*                                                                            */
/* Plugin Parser Implementation Function:                                     */
/*                                                                            */
/******************************************************************************/
void cpiNextParserClassName(CciParser *parser, CciContext *context, CciChar *buffer, int size)
{
	PARSER_CONTEXT_ST *pc = (PARSER_CONTEXT_ST*)context;								/* 17/07/2001 */
	int i = 0;																						/* 23/07/2001 */

#ifdef _DEBUG  
   DBGPRINTF((OST, "->cpiNextParserClassName()\n"));
#endif

   //wcsncpy(buffer, L"none", size);
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
	if (pc->Class[0] != chSpace)																/* 21/07/2001 */
		CciCharNCpy(buffer, pc->Class, size);												/* 21/07/2001 */
	else																								/* 21/07/2001 */
		CciCharNCpy(buffer, constNone, size);												/* 21/07/2001 */
#else																									/* 21/07/2001 */
	if (pc->Class[0] != L' ') {																/* 17/07/2001 */
		wcsncpy(buffer, pc->Class, size);													/* 17/07/2001 */
		for (i = 0; i < size; i++) {															/* 23/07/2001 */
			if (buffer[i] == L' ') {															/* 23/07/2001 */
				buffer[i] = 0;																		/* 23/07/2001 */
				break;																				/* 23/07/2001 */
			}																							/* 23/07/2001 */
		}																								/* 23/07/2001 */
	}																									/* 23/07/2001 */
	else																								/* 17/07/2001 */
		wcsncpy(buffer, L"none", size);														/* 17/07/2001 */
#endif																								/* 21/07/2001 */

#ifdef _DEBUG  
   DBGPRINTF((OST, "<-cpiNextParserClassName(), buffer = %S\n", buffer));
#endif

   return;
}


/******************************************************************************/
/*                                                                            */
/* Plugin Parser Implementation Function:                                     */
/*                                                                            */
/******************************************************************************/
void cpiSetNextParserClassName(CciParser *parser, CciContext *context, CciChar *name, CciBool ParserType)
{
   PARSER_CONTEXT_ST *pc = (PARSER_CONTEXT_ST*)context;

#ifdef _DEBUG  
   DBGPRINTF((OST, "->cpiSetNextParserClassName() name = %S, ParserType = %d\n", name, ParserType));
#endif

   /* Save the name in my context */
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
   CciCharNCpy(pc->NextParserClassName, name, CciCharLen(name));
#else
   wcsncpy(pc->NextParserClassName, name, wcslen(name));
#endif
   /* Save the type in my context */
   pc->ParserType = ParserType;

#ifdef _DEBUG  
   DBGPRINTF((OST, "<-cpiSetNextParserClassName()\n"));
#endif

  return;
}

/******************************************************************************/
/*                                                                            */
/* Plugin Parser Implementation Function:                                     */
/*                                                                            */
/******************************************************************************/
CciBool cpiParserType(CciParser *parser, CciContext *context )
{

#ifdef _DEBUG  
   DBGPRINTF((OST, "<->cpiParserType(), retvalue = 0\n"));
#endif
   /* Is our parser capable of parsing MQ headers? */
   return 0; /* Yes it is */
}


int SetElementValue(CciElement *elt, CciChar *value, int valType)
{
   int err = 0;

   long longVal = 0;

   CciChar p[3];
   CciChar *badChars = NULL;

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
	const char *mbData = NULL;
#endif

   struct CciDate Date; // {int year; int month; int day;};
   struct CciTime Time; // {int hour; int minute; float second;};

#ifdef _DEBUG  
   DBGPRINTF((OST, "->SetElementValue() elt = 0x%x, value = %S, valType = %d\n", elt, value, valType));
#endif
     
   switch (valType) {
      case CCI_VALUE_TYPE_INTEGER:
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
			mbData = mbString(value);			
			longVal = strtol(mbData, (char **)NULL, 10);
			free((void *)mbData);
#else
	      longVal = wcstol(value, &badChars, 10);
#endif
         cpiSetElementIntegerValue(&err, elt, longVal);
         break;
      case CCI_VALUE_TYPE_DATE:
         /* p[2] = L'\0'; */
			p[2] = 0;
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
			CciCharNCpy(p, &value[4], 2);
#else
         wcsncpy(p, &value[4], 2);
#endif

#ifdef WIN32																						/* 22/07/2001 */
         Date.day   = _wtoi(p);
#elif defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
			mbData = mbString(p);																/* 22/07/2001 */
			Date.day = atoi(mbData);															/* 22/07/2001 */
			free((void *)mbData);																/* 22/07/2001 */
#else																									/* 22/07/2001 */
			Date.day   = watoi(p);																/* 22/07/2001 */
#endif																								/* 22/07/2001 */

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
			CciCharNCpy(p, &value[2], 2);
#else
         wcsncpy(p, &value[2], 2);
#endif
#ifdef WIN32																						/* 22/07/2001 */
         Date.month = _wtoi(p);
#elif defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
			mbData = mbString(p);																/* 22/07/2001 */
			Date.month = atoi(mbData);															/* 22/07/2001 */
			free((void *)mbData);																/* 22/07/2001 */
#else																									/* 22/07/2001 */
         Date.month = watoi(p);																/* 22/07/2001 */
#endif																								/* 22/07/2001 */
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
			CciCharNCpy(p, value, 2);
#else
         wcsncpy(p, value, 2);
#endif
#ifdef WIN32																						/* 22/07/2001 */
         Date.year  = 2000 + _wtoi(p);
#elif defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
			mbData = mbString(p);																/* 22/07/2001 */
			Date.year = 2000 + atoi(mbData);													/* 22/07/2001 */
			free((void *)mbData);																/* 22/07/2001 */
#else																									/* 22/07/2001 */
         Date.year  = 2000 + watoi(p);
#endif																								/* 22/07/2001 */
         cpiSetElementDateValue(&err, elt, &Date);
         break;
      case CCI_VALUE_TYPE_TIME:
         /* p[2] = L'\0'; */
			p[2] = 0;
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
			CciCharNCpy(p, value, 2);
#else
         wcsncpy(p, value, 2);
#endif
#ifdef WIN32																						/* 22/07/2001 */
         Time.hour = _wtoi(p);
#elif defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
			mbData = mbString(p);																/* 22/07/2001 */
			Time.hour = atoi(mbData);															/* 22/07/2001 */
			free((void *)mbData);																/* 22/07/2001 */
#else																									/* 22/07/2001 */
			Time.hour = watoi(p);
#endif																								/* 22/07/2001 */

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
			CciCharNCpy(p, &value[2], 2);
#else
         wcsncpy(p, &value[2], 2);
#endif
#ifdef WIN32																						/* 22/07/2001 */
         Time.minute = _wtoi(p);
#elif defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
			mbData = mbString(p);																/* 22/07/2001 */
			Time.minute = atoi(mbData);														/* 22/07/2001 */
			free((void *)mbData);																/* 22/07/2001 */
#else																									/* 22/07/2001 */
         Time.minute = watoi(p);
#endif																								/* 22/07/2001 */

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
			CciCharNCpy(p, &value[4], 2);
#else
         wcsncpy(p, &value[4], 2);
#endif
#ifdef WIN32																						/* 22/07/2001 */
         Time.second = (float)_wtoi(p);
#elif defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
			mbData = mbString(p);																/* 22/07/2001 */
			Time.second = (float)atoi(mbData);												/* 22/07/2001 */
			free((void *)mbData);																/* 22/07/2001 */
#else																									/* 22/07/2001 */
         Time.second = (float)watoi(p);
#endif																								/* 22/07/2001 */

         cpiSetElementTimeValue(&err, elt, &Time);
         break;
      case CCI_VALUE_TYPE_CHAR:
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
			cpiSetElementCharacterValue(&err, elt, value, CciCharLen(value));
#else
	      cpiSetElementCharacterValue(&err, elt, value, wcslen(value));
#endif
	      break;
   }

#ifdef _DEBUG  
   DBGPRINTF((OST, "<-SetElementValue() retvalue = %d\n", err));
#endif

   return err;
}


/******************************************************************************/
/*                                                                            */
/* Plugin Parser Utility Function: Parse the next item                        */
/*                                                                            */
/******************************************************************************/
void *parseNextItem(CciParser *parser, CciContext *context, CciElement *element )
{
   PARSER_CONTEXT_ST *pc = (CciContext *)context;

   CciElement  *returnElement = element;
   CciElement  *newElement    = NULL;

   size_t markedSize;

   void  *endMarker     = NULL;
   void  *startMarker   = NULL;

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
  const CciChar*     data;
#else
  CciChar            data[256];
#endif

   int rc    = -1;
   int vType = -1;

	char CLASSfound = FALSE;

#ifdef _DEBUG    	
   DBGPRINTF((OST, "->parseNextItem() parser = 0x%x, context = 0x%x, element = 0x%x, index = %ld\n", parser, context, element, pc->iIndex));
#endif

   if (pc->iIndex == pc->iSize) {
#ifdef _DEBUG  
      DBGPRINTF((OST, "-->End of buffer reached...\n"));
      DBGPRINTF((OST, "<-parseNextItem() retvalue = 0\n"));
#endif
      return 0;
   }

   /* Skip any white space */
   skipWhiteSpace(context, parser);
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
	if (pc->iCurrentCharacter == chSemiColon) {  
#else
   if (pc->iCurrentCharacter == L';') {
#endif
      advance(context, parser);
      cpiSetElementCompletePrevious(&rc, element, 1);
      cpiSetElementCompleteNext(&rc, element, 1);
#ifdef _DEBUG  
      DBGPRINTF((OST, "-->Set completePrevious/Next on element = 0x%x\n", element));
      DBGPRINTF((OST, "<-parseNextItem() retvalue = 0\n"));
#endif
      return 0;
   }

   /* Get a pointer to the start of the tag */
   startMarker = (char*)pc->iBuffer+(int)pc->iIndex;

   /* Skip over the tag */
   goToNameEnd(context, parser);

   /* Get a pointer to the end of the tag */
   endMarker = (char*)pc->iBuffer+(int)pc->iIndex;

   /* Compute the size of the tag */
   markedSize = (size_t)endMarker-(int)startMarker;

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
   /* Copy the tag into a CciChar form */
   data = CciNString((char *)startMarker, markedSize);
#else
   /* Copy the tag into a wide string */
   memset(&data, 0, sizeof(data));
   mbstowcs(data, startMarker, markedSize);
#endif
   /* Create a new name element for the tag */
   newElement = cpiCreateElement(&rc, parser);
   
   cpiSetElementType(&rc, newElement, CCI_ELEMENT_TYPE_NAME_VALUE); 
   cpiSetElementName(&rc, newElement, data);

   vType = WhatType(data);

#ifdef _DEBUG  
   DBGPRINTF((OST, "-->New keyword found ...\n"));
   DBGPRINTF((OST, "-->Created new NAME element = 0x%x, type = 0x%x, name = %S, size = %ld, index = %ld\n", newElement, CCI_ELEMENT_TYPE_NAME, data, markedSize, pc->iIndex));
#endif          

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
	if (CciCharCmp(data, (CciChar *)constCLASS) == 0) {								/* 21/07/2001 */
#else																									/* 21/07/2001 */
	if (wcscmp(data, L"CLASS" ) == 0) {														/* 17/07/2001 */
#endif																								/* 21/07/2001 */
#ifdef _DEBUG																						/* 17/07/2001 */
		DBGPRINTF((OST, "-->keyword = CLASS\n"));											/* 17/07/2001 */
#endif																								/* 17/07/2001 */
		CLASSfound = TRUE;																		/* 17/07/2001 */
	} /* if (wcscmp(data, L"CLASS" ) == 0) */												/* 17/07/2001 */

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
   /* Free the memory allocated in CciNString() */
   free((void *)data);
#endif

   /* Add the element */
   cpiAddAsLastChild(&rc, element, newElement);
   cpiSetElementCompletePrevious(&rc, newElement, 1);
   cpiSetElementCompleteNext(&rc, newElement, 1);

#ifdef _DEBUG  
   DBGPRINTF((OST, "-->Added 0x%x as lastChild of 0x%x\n", newElement, element));
   DBGPRINTF((OST, "-->Set completePrevious/Next on element = 0x%x\n", newElement));
#endif

   advance(context, parser);

   /* Get a pointer to the start of the text */
   startMarker = (char*)pc->iBuffer + (int)pc->iIndex;

   /* Skip over the text */
   goToValueEnd(context, parser);

   /* Get a pointer to the end of the text */
   endMarker = (char*)pc->iBuffer + (int)pc->iIndex;

   /* Compute the size of the text */
   markedSize = (size_t)endMarker - (int)startMarker;

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
   /* Copy the tag into a CciChar form */
   data = CciNString((char *)startMarker, markedSize);
#else
   /* Copy the tag into a wide string */
   memset(&data, 0, sizeof(data));
   mbstowcs(data, startMarker, markedSize);
#endif
  
   /* Set a new value element for the value */
   SetElementValue(newElement, (CciChar *)data, vType);
   
#ifdef _DEBUG  
   DBGPRINTF((OST, "-->Value found\n"));
   DBGPRINTF((OST, "-->Created new VALUE element = 0x%x, type = 0x%x, name = %S, size = %ld, index = %ld\n", newElement, CCI_ELEMENT_TYPE_VALUE, data, markedSize, pc->iIndex));
#endif

	if (CLASSfound) {																				/* 17/07/2001 */
#ifdef _DEBUG																						/* 17/07/2001 */
		DBGPRINTF((OST, "-->pc->Class = [%S]\n", data));								/* 17/07/2001 */
#endif																								/* 17/07/2001 */
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
		CciCharCpy(pc->Class, data);
#else
		wcscpy(pc->Class, data);																/* 17/07/2001 */
#endif
		CLASSfound = FALSE;																		/* 17/07/2001 */
	} /*	if (CLASSfound) */																	/* 17/07/2001 */

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
   /* Free the memory allocated in CciNString() */
   free((void *)data);
#endif

   advance(context, parser);

#ifdef _DEBUG          
   DBGPRINTF((OST, "<-parseNextItem() retvalue = 0x%x\n", returnElement));
#endif

   return(returnElement);
}


/******************************************************************************/
/*                                                                            */
/* Plugin Parser Implementation Function:                                     */
/*                                                                            */
/******************************************************************************/
int cpiParseBuffer(CciParser *parser, CciContext *context,int offset)
{
   PARSER_CONTEXT_ST *pc = context;

   int rc = -1;

#ifdef _DEBUG  
   DBGPRINTF((OST, "->cpiParseBuffer() parser = 0x%x, context = 0x%x, offset = %d\n", parser, context, offset));
#endif
  
   /* Get a pointer to the message buffer and save the current offset */
   pc->iBuffer = (void *)cpiBufferPointer(&rc, parser);

   if (pc->iBuffer)
      pc->iIndex = offset;
   else 
      pc->iIndex = 0;

   /* Save size of the buffer */
   pc->iSize = cpiBufferSize(&rc, parser);

   /* Prime the first byte in the stream */
   pc->iCurrentCharacter = cpiBufferByte(&rc, parser, pc->iIndex);

   /* Set the current element to the root element */
   pc->iCurrentElement = cpiRootElement(&rc, parser);

	/* Set Class to None */
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
	pc->Class[0] = chSpace;																		/* 21/07/2001 */
#else																									/* 21/07/2001 */
	pc->Class[0] = L' ';																			/* 17/07/2001 */
#endif																								/* 21/07/2001 */

#ifdef _DEBUG
   DBGPRINTF((OST, "->cpiParseBuffer() iCurrentElement = 0x%x\n", pc->iCurrentElement));
#endif

   while (pc->iCurrentElement)
      pc->iCurrentElement = parseNextItem(parser, context, pc->iCurrentElement);
  
#ifdef _DEBUG  
   DBGPRINTF((OST, "<-cpiParseBuffer() pc->iIndex = retvalue = %d\n", pc->iIndex));
#endif

  return(pc->iIndex);
}


int WriteElementValue(NodeStringBuffer *buf, CciElement *elt)
{
   int err       = 0;

   char tmpBuf[64];
   char *str = NULL;

   CciSize outBufLen = INITIAL_BUF_LEN;

   CciValueType type;
   
   CciChar *outBuf = NULL;

   struct CciDate Date;
   struct CciTime Time;

#ifdef _DEBUG
   DBGPRINTF((OST, "->WriteElementValue element = 0x%x\n", elt));
#endif

   type = cniElementValueType(&err, elt);

   outBuf = (CciChar *)malloc(sizeof(CciChar) * outBufLen);

   switch(type) {
      case CCI_VALUE_TYPE_BOOL:
	      if (cpiElementBooleanValue(&err, elt))
	         NodeStringBufferAdd(buf, "TRUE");
	      else
	         NodeStringBufferAdd(buf, "FALSE");
	      break;
      case CCI_VALUE_TYPE_INTEGER:
	      sprintf(tmpBuf, "%Li", cpiElementIntegerValue(&err, elt));
	      if (err == 0) 
	         NodeStringBufferAdd(buf, tmpBuf);
	      break;
      case CCI_VALUE_TYPE_REAL:
	      sprintf(tmpBuf, "%16.16lf", cpiElementRealValue(&err, elt));
	      if (err == 0)
	         NodeStringBufferAdd(buf, tmpBuf);
		   break;
      case CCI_VALUE_TYPE_DECIMAL:
	      cpiElementDecimalValue(&err, elt, outBuf, outBufLen);
	      if (err == 0) {
	         str = UnicodeToASCII(outBuf);
	         NodeStringBufferAdd(buf, str);
	         free(str);
	      }
	      break;
      case CCI_VALUE_TYPE_CHAR:
	      do {
	         cpiElementCharacterValue(&err, elt, outBuf, outBufLen);
	         if (err == CCI_BUFFER_TOO_SMALL) {
		         outBufLen *= 2;
		         outBuf = (CciChar *)realloc(outBuf, sizeof(CciChar) * outBufLen);
	         }
	      } while (err == CCI_BUFFER_TOO_SMALL);

	      if (err == 0) {
	         str = UnicodeToASCII(outBuf);
	         NodeStringBufferAdd(buf, str);
	         free(str);
         }
	      break;
      case CCI_VALUE_TYPE_DATE:
         Date = cpiElementDateValue(&err, elt);
         if (err == 0) {
            sprintf(tmpBuf, "%.2d%.2d%.2d", Date.year - 2000, Date.month, Date.day);
            NodeStringBufferAdd(buf, tmpBuf);
         }
         else
            NodeStringBufferAdd(buf, "<date>");
         break;
      case CCI_VALUE_TYPE_TIME:
      case CCI_VALUE_TYPE_GMTTIME:
         Time = cpiElementTimeValue(&err, elt);
         if (err == 0) {
            sprintf(tmpBuf, "%.2d%.2d%.2d", Time.hour, Time.minute, (int)Time.second);
            NodeStringBufferAdd(buf, tmpBuf);
         }
         else
            NodeStringBufferAdd(buf, "<time>");
	      break;
      case CCI_VALUE_TYPE_TIMESTAMP:
      case CCI_VALUE_TYPE_GMTTIMESTAMP:
	      NodeStringBufferAdd(buf, "<timestamp>");
	      break;
      case CCI_VALUE_TYPE_BYTEARRAY:
      case CCI_VALUE_TYPE_BITARRAY:
	      NodeStringBufferAdd(buf, "<binary>");
	      break;
      default:
	      NodeStringBufferAdd(buf, "<unknown>");
	      break;
   } /* switch(type) */

   free(outBuf);

#ifdef _DEBUG
   DBGPRINTF((OST, "<-WriteElementValue retvalue = %d\n", err));
#endif

   return err;
}


int WriteElement(NodeStringBuffer *buf, CciElement *elt, char *eltName)
{
   int err = 0;

#ifdef _DEBUG
   DBGPRINTF((OST, "->WriteElement element = 0x%x, eltName = %s\n", elt, eltName));
#endif

   if ((elt != NULL) && (eltName != NULL)) {
      NodeStringBufferAdd(buf, eltName);
      NodeStringBufferAddChar(buf, '(');

      err = WriteElementValue(buf, elt);

      NodeStringBufferAddChar(buf, ')');
      NodeStringBufferAddChar(buf, ' ');
   } /* if ((elt != NULL) && (eltName != NULL)) */

#ifdef _DEBUG
   DBGPRINTF((OST, "<-WriteElement retvalue = %d\n", err));
#endif

   return err;
}


int cpiWriteBuffer(CciParser *parser, CciContext *context)
{
   int err    = -1;
   int retVal = -1;

   NodeStringBuffer *buf = NULL;

   CciElement *rootElt    = NULL;
   CciElement *curElement = NULL;

   char *aStr = NULL;

   CciChar eltName[INITIAL_BUF_LEN];

#ifdef _DEBUG
   DBGPRINTF((OST, "->cpiWriteBuffer\n"));
#endif

   rootElt = cpiRootElement(&err, parser);

   buf = NodeStringBufferAlloc();

   /* Output all of our children */
   curElement = cpiFirstChild(&err, rootElt);

   while (curElement != NULL && (err == 0)) {
      cpiElementName(&err, curElement, eltName, INITIAL_BUF_LEN);
	   aStr = UnicodeToASCII(eltName);
      err = WriteElement(buf, curElement, aStr);
      free(aStr);
	   curElement = cpiNextSibling(&err, curElement);
   }

   NodeStringBufferAddChar(buf, ';');

   /* We're done -- write out the final buffer to the message */
   cpiAppendToBuffer(&err, parser, buf->buf, buf->length);

   retVal = buf->length;
   NodeStringBufferFree(buf);

#ifdef _DEBUG
   DBGPRINTF((OST, "<-cpiWriteBuffer retvalue = %d\n", retVal));
#endif
   return retVal;
}


#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */
/*  initParserConstants                                                       */
/*                                                                            */
/*  Creates Cci constant strings from defined char * constant strings.        */
/*                                                                            */
/******************************************************************************/
void initParserConstants(void)
{
	constMQSIE				= (CciChar *)CciString(CONST_MQSIE);
	constParserFactory	= (CciChar *)CciString(CONST_PARSER_FACTORY);

	constNone				= (CciChar *)CciString(CONST_NONE);

	constDATATYPE			= (CciChar *)CciString(CONST_DATATYPE);
	constCLASS				= (CciChar *)CciString(CONST_CLASS);
	constCONTROLNUM		= (CciChar *)CciString(CONST_CONTROLNUM);
	constMSGNAME			= (CciChar *)CciString(CONST_MSGNAME);
	constMSGSEQNO			= (CciChar *)CciString(CONST_MSGSEQNO);
	constRECLEN				= (CciChar *)CciString(CONST_RECLEN);
	constRECDLM				= (CciChar *)CciString(CONST_RECDLM);
	constRECFM				= (CciChar *)CciString(CONST_RECFM);
	constSENDER				= (CciChar *)CciString(CONST_SENDER);
	constSENDQUAL			= (CciChar *)CciString(CONST_SENDQUAL);
	constRECEIVER			= (CciChar *)CciString(CONST_RECEIVER);
	constRECVQUAL			= (CciChar *)CciString(CONST_RECVQUAL);
	constFROM				= (CciChar *)CciString(CONST_FROM);
	constUNIQUEID			= (CciChar *)CciString(CONST_UNIQUEID);
	constDATE				= (CciChar *)CciString(CONST_DATE);
	constTIME				= (CciChar *)CciString(CONST_TIME);
}
#endif


void LilFactoryExportPrefix *LilFactoryExportSuffix bipGetParserFactory()
{
  /* Declare variables */
  CciFactory*     factoryObject;
  int             rc;
  static CPI_VFT  vftable;

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
  /* Before we proceed we need to initialise all the static constants */
  /* that may be used by the plug-in.                                 */
  initParserConstants();
#endif	
  /* Setup function table with pointers to parser implementation functions */
  memset(&vftable, 0, sizeof(vftable));
  vftable.iFpCreateContext            = cpiCreateContext;
  vftable.iFpParseBuffer              = cpiParseBuffer;
  vftable.iFpParseFirstChild          = cpiParseFirstChild;
  vftable.iFpParseLastChild           = cpiParseLastChild;
  vftable.iFpParseLeftSibling         = cpiParseLeftSibling;
  vftable.iFpParseRightSibling        = cpiParseRightSibling;
  vftable.iFpWriteBuffer              = cpiWriteBuffer;
  vftable.iFpDeleteContext            = cpiDeleteContext;
  vftable.iFpSetValue                 = 0;//cpiSetValue;
  vftable.iFpValue                    = 0;//cpiValue;
  vftable.iFpNextParserClassName      = cpiNextParserClassName;
  vftable.iFpSetNextParserClassName   = cpiSetNextParserClassName;
  vftable.iFpNextParserEncoding       = 0;//cpiNextParserEncoding;
  vftable.iFpNextParserCodedCharSetId = 0;//cpiNextParserCodedCharSetId;
  vftable.iFpIsHeaderParser           = cpiParserType;

  /* Create the parser factory for this plugin */
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
  factoryObject = cpiCreateParserFactory(&rc, constParserFactory); 
#else
  factoryObject = cpiCreateParserFactory(&rc, L"ComIbmMQSIEParserFactory");
#endif
  if (factoryObject) {
    /* Define the classes of message supported by the factory */
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
    cpiDefineParserClass(&rc, factoryObject, constMQSIE, &vftable);
#else
    cpiDefineParserClass(&rc, factoryObject, L"MQSIE", &vftable);
#endif
  }
  else {
    /* Error: Unable to create parser factory */ 
  }

  /* Return address of this factory object to the broker */
  return(factoryObject);
}

